@@ -1,7 +1,7 @@
 /*
  *  Copyright (C) 2004 Adam Hooper
- *  Copyright (C) 2004 Jean-François Rameau
- *  Copyright (C) 2004 Christian Persch
+ *  Copyright (C) 2004, 2005 Jean-François Rameau
+ *  Copyright (C) 2004, 2005 Christian Persch
  *
  *  Ripped from GaleonWrapper.cpp, which has no copyright info besides Marco...
  *
@@ -19,7 +19,7 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
- *  $Id: mozilla-helpers.cpp,v 1.23 2005/01/13 21:43:46 jframeau Exp $
+ *  $Id: mozilla-helpers.cpp,v 1.31 2005/04/26 12:03:40 chpe Exp $
  */
 
 #include "mozilla-config.h"
@@ -27,6 +27,7 @@
 #include "config.h"
 
 #include "mozilla-helpers.h"
+#include "PageInfoPrivate.h"
 
 #include "EphyUtils.h"
 
@@ -35,9 +36,9 @@
 #include <gtkmozembed.h>
 #include <gtkmozembed_internal.h>
 
-#define MOZILLA_STRICT_API
+#undef MOZILLA_INTERNAL_API
 #include <nsEmbedString.h>
-#undef MOZILLA_STRICT_API
+#define MOZILLA_INTERNAL_API 1
 #include <nsCOMPtr.h>
 
 #include <nsCompatibility.h>
@@ -67,6 +68,7 @@
 #include <nsIDOMHTMLImageElement.h>
 #include <nsIDOMHTMLInputElement.h>
 #include <nsIDOMHTMLLinkElement.h>
+#include <nsIDOMHTMLMetaElement.h>
 #include <nsIDOMHTMLObjectElement.h>
 #include <nsIDOMLocation.h>
 #include <nsIDOMNodeFilter.h>
@@ -77,7 +79,6 @@
 #include <nsIDOMTreeWalker.h>
 #include <nsIDOMViewCSS.h>
 #include <nsIDOMWindow.h>
-#include <nsIHTMLDocument.h>
 #include <nsIInterfaceRequestorUtils.h>
 #include <nsIPrefBranch.h>
 #include <nsIPrefService.h>
@@ -112,9 +113,12 @@ private:
                     nsIDOMHTMLElement *aHTMLElement);
   void ProcessAppletNode (nsIDOMHTMLAppletElement *aElement);
   void ProcessAreaNode (nsIDOMHTMLAreaElement *aElement);
+  void PageInfoHelper::ProcessEmbedNodeHelper (const nsEmbedString &aUrl, 
+  				               nsIDOMHTMLEmbedElement *aElement);
   void ProcessEmbedNode (nsIDOMHTMLEmbedElement *aElement);
   void ProcessFormNode (nsIDOMHTMLFormElement *aElement);
   void ProcessImageNode (nsIDOMHTMLImageElement *aElement);
+  void ProcessMetaNode (nsIDOMHTMLMetaElement *aElement);
   void ProcessInputNode (nsIDOMHTMLInputElement *aElement);
   template <class T> void ProcessLinkNode (nsIDOMNode *aNode);
   void ProcessObjectNode (nsIDOMHTMLObjectElement *aElement);
@@ -136,6 +140,7 @@ private:
   GHashTable *mMediaHash;
   GHashTable *mLinkHash;
   GHashTable *mFormHash;
+  GList      *mMetaTagsList;
   
   /* these two are used by the walker: */
   nsEmbedCString mDocCharset;
@@ -144,9 +149,11 @@ private:
 
 PageInfoHelper::PageInfoHelper ()
 {
-  mMediaHash = g_hash_table_new (g_str_hash, g_str_equal);
-  mLinkHash = g_hash_table_new (g_str_hash, g_str_equal);
-  mFormHash = g_hash_table_new (g_str_hash, g_str_equal);
+  mMediaHash     = g_hash_table_new (g_str_hash, g_str_equal);
+  mLinkHash      = g_hash_table_new (g_str_hash, g_str_equal);
+  mFormHash      = g_hash_table_new (g_str_hash, g_str_equal);
+  /* list of (name,content) */
+  mMetaTagsList  = NULL;
 }
 
 PageInfoHelper::~PageInfoHelper ()
@@ -211,6 +218,77 @@ PageInfoHelper::Init (EphyEmbed *aEmbed)
   return NS_OK;
 }
 
+/* A fairly good case insensitive comparison */
+static gint 
+utf8_strcasecmp (const gchar *a, const gchar *b)
+{
+	gint retval;
+	char *str_a, *str_b;
+
+	str_a = g_utf8_casefold (a, -1);
+	str_b = g_utf8_casefold (b, -1);
+	retval = g_utf8_collate (str_a, str_b);
+	g_free (str_a);
+	g_free (str_b);
+
+	return retval;
+}
+
+/* Compares two links */
+static gint 
+link_compare (gconstpointer a, 
+	      gconstpointer b)
+{
+	EmbedPageLink *link_a = (EmbedPageLink *)a;
+	EmbedPageLink *link_b = (EmbedPageLink *)b;
+
+	return utf8_strcasecmp (link_a->url, link_b->url);
+}
+
+/* Compares two media, first by type then by url */
+static gint 
+media_compare (gconstpointer a, 
+	       gconstpointer b)
+{
+	gint res;
+
+	EmbedPageMedium *media_a = (EmbedPageMedium *)a;
+	EmbedPageMedium *media_b = (EmbedPageMedium *)b;
+
+	// Same type then compare urls.
+	if (media_a->type == media_b->type)
+	{
+		res = utf8_strcasecmp (media_a->url, media_b->url);
+	}
+	else
+	{
+		// Enum comparison ...
+		res = media_a->type > media_b->type ? 1 : -1;
+   	}
+
+	return res;
+}
+
+/* Compares two forms */
+static gint 
+form_compare (gconstpointer a, 
+	      gconstpointer b)
+{
+	EmbedPageForm *form_a = (EmbedPageForm *)a;
+	EmbedPageForm *form_b = (EmbedPageForm *)b;
+
+	// be careful, form's name may be NULL	
+	if (form_a->name == NULL)
+	{
+		return form_b->name == NULL ? 0 : -1; 
+	}
+	else if (form_b->name == NULL)
+	{
+		return form_a->name == NULL ? 0 : 1; 
+	}
+	return utf8_strcasecmp (form_a->name, form_b->name);
+}
+
 static void
 make_list (gpointer key,
            gpointer value,
@@ -229,9 +307,17 @@ PageInfoHelper::GetInfo ()
   EmbedPageInfo *info = g_new0 (EmbedPageInfo, 1);
 
   info->props = GetProperties ();
+
   g_hash_table_foreach (mMediaHash, (GHFunc) make_list, &info->media);
+  info->media = g_list_sort (info->media, media_compare);
+
   g_hash_table_foreach (mLinkHash, (GHFunc) make_list, &info->links);
+  info->links = g_list_sort (info->links, link_compare);
+
   g_hash_table_foreach (mFormHash, (GHFunc) make_list, &info->forms);
+  info->forms = g_list_sort (info->forms, form_compare);
+
+  info->metatags = mMetaTagsList;
 
   return info;
 }
@@ -277,7 +363,7 @@ PageInfoHelper::GetCacheEntryDescriptor (const nsAString &aUrl,
   
       nsCOMPtr<nsICacheEntryDescriptor> cacheEntryDescriptor;
   
-#ifdef MOZ_NSICACHESESSION_OPENCACHEENTRY_NSACSTRING_
+#ifdef HAVE_GECKO_1_8
       rv = cacheSession->OpenCacheEntry (nsEmbedCString(url), nsICache::ACCESS_READ,
                                          PR_FALSE, aEntry);
 #else
@@ -368,12 +454,8 @@ PageInfoHelper::GetProperties ()
 	}
     }
 
-  /* Might not work on XUL pages */
-  nsCOMPtr<nsIHTMLDocument> htmlDoc (do_QueryInterface (mDOMDocument));
-  if (htmlDoc)
-    {
-      props->rendering_mode = (EmbedPageRenderMode) htmlDoc->GetCompatibilityMode();
-    }
+  /* Until https://bugzilla.mozilla.org/show_bug.cgi?id=154359 is fixed */
+  props->rendering_mode = PageInfoPrivate::GetRenderMode (mDOMDocument);
 
   /* Get the URL so we can look in the cache for this page */
   nsCOMPtr<nsIDOMLocation> domLocation;
@@ -519,24 +601,25 @@ PageInfoHelper::ProcessAreaNode (nsIDOMHTMLAreaElement *aElement)
 }
 
 void
-PageInfoHelper::ProcessEmbedNode (nsIDOMHTMLEmbedElement *aElement)
+PageInfoHelper::ProcessEmbedNodeHelper (const nsEmbedString &aUrl, 
+  				        nsIDOMHTMLEmbedElement *aElement)
 {
   nsresult rv;
-  nsEmbedString tmp;
-  rv = aElement->GetSrc (tmp);
-  if (NS_FAILED (rv) || !tmp.Length()) return;
 
   nsEmbedCString cUrl;
-  rv = Resolve (tmp, cUrl);
+  rv = Resolve (aUrl, cUrl);
   if (NS_FAILED (rv) || !cUrl.Length()) return;
 
   if (g_hash_table_lookup (mMediaHash, cUrl.get())) return;
 
-  EmbedPageMedium *medium = g_new0 (EmbedPageMedium, 1);
+  EmbedPageMedium *medium;
+
+  medium = g_new0 (EmbedPageMedium, 1);
   medium->type = MEDIUM_EMBED;
   medium->url = g_strdup (cUrl.get());
   g_hash_table_insert (mMediaHash, medium->url, medium);
 
+  nsEmbedString tmp;
   rv = aElement->GetTitle (tmp);
   if (NS_SUCCEEDED (rv))
     {
@@ -544,6 +627,27 @@ PageInfoHelper::ProcessEmbedNode (nsIDOMHTMLEmbedElement *aElement)
     }
 }
 
+void
+PageInfoHelper::ProcessEmbedNode (nsIDOMHTMLEmbedElement *aElement)
+{
+  nsresult rv;
+  nsEmbedString tmp;
+
+  // Look at src tag
+  rv = aElement->GetSrc (tmp);
+  if (NS_SUCCEEDED (rv) && tmp.Length())
+    {
+      ProcessEmbedNodeHelper (tmp, aElement);
+    }
+
+  // Look at href tag
+  aElement->GetAttribute(mHrefAttr, tmp);
+  if (NS_SUCCEEDED (rv) && tmp.Length())
+    { 
+      ProcessEmbedNodeHelper (tmp, aElement);
+    }
+}
+
 void
 PageInfoHelper::ProcessFormNode (nsIDOMHTMLFormElement *aElement)
 {
@@ -635,6 +739,30 @@ PageInfoHelper::ProcessObjectNode (nsIDOMHTMLObjectElement *aElement)
     }
 }
 
+void
+PageInfoHelper::ProcessMetaNode (nsIDOMHTMLMetaElement *aElement)
+{
+  nsresult rv;
+  nsEmbedString key;
+  /* if we have a http-equiv in the page */
+  rv = aElement->GetHttpEquiv (key);
+  if (NS_FAILED (rv) || !key.Length())
+  {
+    rv = aElement->GetName (key);
+    if (NS_FAILED (rv) || !key.Length()) return;
+  }
+
+  nsEmbedString content;
+  rv = aElement->GetContent (content);
+  if (NS_FAILED (rv) || !content.Length ()) return;
+
+  EmbedPageMetaTag *tag = g_new0 (EmbedPageMetaTag, 1);
+  tag->name = ToCString (key);
+  tag->content =  ToCString (content);
+
+  mMetaTagsList = g_list_prepend (mMetaTagsList, tag);
+}
+
 void
 PageInfoHelper::ProcessInputNode (nsIDOMHTMLInputElement *aElement)
 {
@@ -795,8 +923,11 @@ PageInfoHelper::WalkTree (nsIDOMDocument *aDocument)
   rv = EphyUtils::NewURI (getter_AddRefs (mBaseURI), spec, mDocCharset.get(), nsnull);
   if (NS_FAILED (rv) || !mBaseURI) return;
 
+  nsCOMPtr<nsIDOMNode> rootNode (do_QueryInterface (htmlDoc));
+  if (!rootNode) return;
+
   nsCOMPtr<nsIDOMTreeWalker> walker;
-  rv = trav->CreateTreeWalker(htmlDoc,
+  rv = trav->CreateTreeWalker(rootNode,
                               nsIDOMNodeFilter::SHOW_ELEMENT,
                               nsnull,
                               PR_TRUE,
@@ -886,6 +1017,12 @@ PageInfoHelper::WalkTree (nsIDOMDocument *aDocument)
           ProcessFormNode (nodeAsForm);
         }
 
+      nsCOMPtr<nsIDOMHTMLMetaElement> nodeAsMeta (do_QueryInterface(aNode));
+      if (nodeAsMeta)
+        {
+          ProcessMetaNode (nodeAsMeta);
+        }
+
       nsCOMPtr<nsIDOMHTMLInputElement> nodeAsInput (do_QueryInterface(aNode));
       if (nodeAsInput)
         {
@@ -988,6 +1125,14 @@ free_embed_page_form (EmbedPageForm *form)
   g_free (form);
 }
 
+static void
+free_embed_page_metatags (EmbedPageMetaTag *meta)
+{
+  g_free (meta->name);
+  g_free (meta->content);
+  g_free (meta);
+}
+
 static void
 mozilla_free_page_properties (EmbedPageProperties *props)
 {
@@ -1009,6 +1154,9 @@ mozilla_free_embed_page_info (EmbedPageInfo *info)
   g_list_foreach (info->media, (GFunc) free_embed_page_medium, NULL);
   g_list_free (info->media);
 
+  g_list_foreach (info->metatags, (GFunc) free_embed_page_metatags, NULL);
+  g_list_free (info->metatags);
+
   mozilla_free_page_properties (info->props);
 
   g_free (info);
